import { basename, dirname, extname, join } from "node:path";
import type babel from "@babel/core";
import { type NodePath, type types as t } from "@babel/core";
import type { SourceLocation, StyleMap, StyleMapEntry } from "./shared";
import { type TailwindPluginOptions, getClassName } from "./index";

const definePlugin =
  <T>(fn: (runtime: typeof babel) => babel.Visitor<babel.PluginPass & T>) =>
  (runtime: typeof babel) => {
    const plugin: babel.PluginObj<babel.PluginPass & T> = {
      visitor: fn(runtime),
    };
    return plugin as babel.PluginObj;
  };

const extractJSXContainer = (attr: NonNullable<t.JSXAttribute["value"]>): t.Expression =>
  attr.type === "JSXExpressionContainer" ? (attr.expression as t.Expression) : attr;

function matchPath(
  nodePath: NodePath<t.Node | null | undefined>,
  fns: (dig: (nodePath: NodePath) => void) => babel.Visitor
) {
  if (!nodePath.node) return;
  const fn = fns(path => matchPath(path, fns))[nodePath.node.type] as any;
  fn?.(nodePath);
}

interface BabelPluginState {
  getCx: () => t.Identifier;
  sliceText: (node: t.Node) => SourceLocation;
  recordIfAbsent: (node: StyleMapEntry) => void;
  tailwindMap: Map<string, StyleMapEntry>;
}

export function babelTailwind(
  styleMap: StyleMap,
  {
    clsx,
    getClassName: getClass = getClassName,
    taggedTemplateName,
    jsxAttributeAction = "delete",
    jsxAttributeName = "css",
  }: TailwindPluginOptions
) {
  function getClsxImport(t: typeof babel.types, cx: t.Identifier) {
    switch (clsx) {
      case "emotion":
        return t.importDeclaration(
          [t.importSpecifier(cx, t.identifier("cx"))],
          t.stringLiteral("@emotion/css")
        );
      case "clsx":
        return t.importDeclaration(
          [t.importDefaultSpecifier(cx)],
          t.stringLiteral("clsx")
        );
      case "classnames":
        return t.importDeclaration(
          [t.importDefaultSpecifier(cx)],
          t.stringLiteral("classnames")
        );
      default:
        throw new Error("Unknown clsx library");
    }
  }

  return definePlugin<BabelPluginState>(({ types: t }) => ({
    Program: {
      enter(path, state) {
        let cx: t.Identifier;
        const map = (state.tailwindMap = new Map());
        state.sliceText = node => ({
          filename: state.filename!,
          start: node.loc!.start,
          end: node.loc!.end,
          text: state.file.code
            .split("\n")
            .slice(node.loc!.start.line - 1, node.loc!.end.line)
            .join("\n"),
        });
        state.recordIfAbsent = entry => {
          if (!map.has(entry.key)) {
            map.set(entry.key, entry);
          }
        };
        state.getCx = () => {
          if (cx == null) {
            cx = path.scope.generateUidIdentifier("cx");
            path.node.body.unshift(getClsxImport(t, cx));
          }
          return t.cloneNode(cx);
        };
      },

      exit({ node }, { filename, tailwindMap }) {
        if (!tailwindMap.size) return;
        if (!filename) {
          throw new Error("babel: missing state.filename");
        }

        const cssName = basename(filename, extname(filename)) + ".css";
        node.body.unshift(
          t.importDeclaration([], t.stringLiteral(`tailwind:./${cssName}`))
        );

        styleMap.set(join(dirname(filename), cssName), Array.from(tailwindMap.values()));
      },
    },

    TaggedTemplateExpression(path, { sliceText, recordIfAbsent }) {
      if (taggedTemplateName == null) return;
      const { node } = path;

      const {
        tag,
        quasi: { quasis, expressions },
      } = node;
      if (!t.isIdentifier(tag, { name: taggedTemplateName })) return;

      if (expressions.length) {
        throw new Error(`${taggedTemplateName}\`\` should not contain expressions`);
      }

      const value = quasis[0].value.cooked;
      if (value) {
        const trimmed = value.replace(/\s+/g, " ").trim();
        const className = getClass(trimmed);
        recordIfAbsent({
          key: className,
          className: trimmed,
          location: sliceText(node),
        });
        path.replaceWith(t.stringLiteral(className));
      }
    },

    JSXAttribute(path, { sliceText, recordIfAbsent, getCx }) {
      const { name } = path.node;
      if (name.name !== jsxAttributeName) return;

      const valuePath = path.get("value");
      if (!valuePath.node) return;

      const copy =
        jsxAttributeAction === "delete" ? undefined : t.cloneNode(valuePath.node, true);

      const parent = path.parent as t.JSXOpeningElement;
      const classNameAttribute = parent.attributes.find(
        (attr): attr is t.JSXAttribute =>
          t.isJSXAttribute(attr) && attr.name.name === "className"
      );

      matchPath(valuePath, go => ({
        StringLiteral(path) {
          const { node } = path;
          const { value } = node;

          if (value) {
            const trimmed = value.replace(/\s+/g, " ").trim();
            const className = getClass(trimmed);
            recordIfAbsent({
              key: className,
              className: trimmed,
              location: sliceText(node),
            });
            path.replaceWith(t.stringLiteral(className));
          }
        },
        JSXExpressionContainer(path) {
          go(path.get("expression"));
        },
        ConditionalExpression(path) {
          go(path.get("consequent"));
          go(path.get("alternate"));
        },
        LogicalExpression(path) {
          go(path.get("right"));
        },
        CallExpression(path) {
          for (const arg of path.get("arguments")) {
            go(arg);
          }
        },
      }));

      if (classNameAttribute) {
        const attrValue = classNameAttribute.value!;

        // If both are string literals, we can merge them directly here
        if (t.isStringLiteral(attrValue) && t.isStringLiteral(valuePath.node)) {
          attrValue.value +=
            (attrValue.value.at(-1) === " " ? "" : " ") + valuePath.node.value;
        } else {
          classNameAttribute.value = t.jsxExpressionContainer(
            t.callExpression(getCx(), [
              extractJSXContainer(attrValue),
              extractJSXContainer(valuePath.node),
            ])
          );
        }
      } else {
        parent.attributes.push(
          t.jsxAttribute(t.jsxIdentifier("className"), valuePath.node)
        );
      }

      if (jsxAttributeAction === "delete") {
        path.remove();
      } else {
        path.node.value = copy!;
        if (Array.isArray(jsxAttributeAction) && jsxAttributeAction[0] === "rename") {
          path.node.name.name = jsxAttributeAction[1];
        }
      }
    },
  }));
}
