import type { Declaration } from "postcss";
import type { ResolvedTailwindConverterConfig } from "../index";

import UTILITIES_MAPPING from "./utilities-mapping.json";
import {
  normalizeColorValue,
  normalizeSizeValue,
  normalizeValue,
} from "../utils/converterMappingByTailwindTheme";
import { isCSSVariable, removeUnnecessarySpaces } from "../utils";

function parseCSSFunction(string: string) {
  const { name, value } = string.match(/(?<name>[\w-]+)\((?<value>.*?)\)/)?.groups || {};
  return { name: name || null, value: value || null };
}

const parseCSSFunctions = (value: string) =>
  value
    .trim()
    .match(/(?<name>[\w-]+)\((?<value>.*?)\)/gm)
    ?.map((cssFunction: string) => parseCSSFunction(cssFunction)) || [];

export function prepareArbitraryValue(value: string) {
  return normalizeValue(value).replace(/_/g, "\\_").replace(/\s+/g, "_");
}

type CSSDataType = "color" | "length" | "number" | "image" | "position" | "family-name";

export function convertDeclarationValue({
  value,
  valuesMap,
  classPrefix,
  fallbackValue = value,
  fallbackClassPrefix = classPrefix,
  cssDataType = null,
}: {
  value: string;
  valuesMap: Record<string, string>;
  classPrefix: string;
  fallbackValue?: string;
  fallbackClassPrefix?: string;
  cssDataType?: CSSDataType | null;
}) {
  const normalizedValue = normalizeValue(value);
  const mappedValue = valuesMap[normalizedValue];
  if (mappedValue) {
    if (mappedValue === "DEFAULT") {
      return [classPrefix];
    }

    return [`${classPrefix}-${mappedValue}`];
  }

  const arbitraryValue = prepareArbitraryValue(fallbackValue);

  if (!arbitraryValue) {
    return [];
  }

  if (cssDataType && isCSSVariable(arbitraryValue)) {
    return [
      `${fallbackClassPrefix}-[${cssDataType ? `${cssDataType}:` : ""}${arbitraryValue}]`,
    ];
  }

  return [`${fallbackClassPrefix}-[${arbitraryValue}]`];
}

export function strictConvertDeclarationValue(
  value: string,
  valuesMap: Record<string, string>
) {
  return valuesMap[value] ? [valuesMap[value]] : [];
}

function convertColorDeclarationValue({
  declValue,
  valuesMap,
  classPrefix,
  cssDataType = null,
}: {
  declValue: string;
  valuesMap: Record<string, string>;
  classPrefix: string;
  cssDataType?: CSSDataType | null;
}) {
  return convertDeclarationValue({
    value: normalizeColorValue(declValue),
    valuesMap,
    classPrefix,
    fallbackValue: declValue,
    fallbackClassPrefix: classPrefix,
    cssDataType,
  });
}

function convertSizeDeclarationValue({
  declValue,
  valuesMap,
  classPrefix,
  remInPx,
  supportsNegativeValues = false,
  cssDataType = null,
}: {
  declValue: string;
  valuesMap: Record<string, string>;
  classPrefix: string;
  remInPx: number | null | undefined;
  supportsNegativeValues?: boolean;
  cssDataType?: CSSDataType | null;
}) {
  const normalizedValue = normalizeSizeValue(declValue, remInPx);
  const isNegativeValue = supportsNegativeValues && normalizedValue.startsWith("-");

  return convertDeclarationValue({
    value: isNegativeValue ? normalizedValue.slice(1) : normalizedValue,
    valuesMap,
    classPrefix: isNegativeValue ? `-${classPrefix}` : classPrefix,
    fallbackValue: declValue,
    fallbackClassPrefix: classPrefix,
    cssDataType,
  });
}

function convertBorderDeclarationValue({
  value,
  config,
  classPrefix,
}: {
  value: string;
  config: ResolvedTailwindConverterConfig;
  classPrefix: string;
}) {
  const [width, style, ...colorArray] = value.split(/\s+/m);
  const color = colorArray.join(" ");

  let classes: string[] = [];

  if (width) {
    if (!config.tailwindConfig.corePlugins.borderWidth) {
      return [];
    }
    classes = classes.concat(
      convertSizeDeclarationValue({
        declValue: width,
        valuesMap: config.mapping.borderWidth,
        classPrefix,
        remInPx: config.remInPx,
        supportsNegativeValues: false,
        cssDataType: "length",
      })
    );
  }

  if (style) {
    if (!config.tailwindConfig.corePlugins.borderStyle) {
      return [];
    }
    classes = classes.concat(
      strictConvertDeclarationValue(style, UTILITIES_MAPPING["border-style"])
    );
  }

  if (color) {
    if (!config.tailwindConfig.corePlugins.borderColor) {
      return [];
    }
    classes = classes.concat(
      convertColorDeclarationValue({
        declValue: color,
        valuesMap: config.mapping.borderColor,
        classPrefix,
        cssDataType: "color",
      })
    );
  }

  return classes;
}

function parseComposedSpacingValue(value: string) {
  const values = value.split(/\s+/m);

  if (values.length > 4) {
    return { top: null, right: null, bottom: null, left: null };
  }

  return {
    top: values[0],
    right: values[1] || values[0],
    bottom: values[2] || values[0],
    left: values[3] || values[1] || values[0],
  };
}

export function convertComposedSpacingDeclarationValue(
  value: string,
  mapping: {
    top: { valuesMapping: Record<string, string>; classPrefix: string };
    right: { valuesMapping: Record<string, string>; classPrefix: string };
    bottom: { valuesMapping: Record<string, string>; classPrefix: string };
    left: { valuesMapping: Record<string, string>; classPrefix: string };
  },
  remInPx: number | null | undefined
) {
  const parsed = parseComposedSpacingValue(value);
  let classes: string[] = [];

  for (const [key, value] of Object.entries(parsed) as [
    "top" | "right" | "bottom" | "left",
    any,
  ][]) {
    const { valuesMapping, classPrefix } = mapping[key] || {};

    if (value && valuesMapping && classPrefix) {
      classes = classes.concat(
        convertSizeDeclarationValue({
          declValue: value,
          valuesMap: valuesMapping,
          classPrefix,
          remInPx,
          supportsNegativeValues: true,
        })
      );
    }
  }

  return classes;
}

interface DeclarationConvertersMapping {
  [property: string]: (
    declaration: Declaration,
    config: ResolvedTailwindConverterConfig
  ) => string[];
}

export const DECLARATION_CONVERTERS_MAPPING: DeclarationConvertersMapping = {
  "accent-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.accentColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.accentColor,
          classPrefix: "accent",
          cssDataType: "color",
        })
      : [],

  "align-content": (declaration, config) =>
    config.tailwindConfig.corePlugins.alignContent
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["align-content"]
        )
      : [],

  "align-items": (declaration, config) =>
    config.tailwindConfig.corePlugins.alignItems
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["align-items"])
      : [],

  "align-self": (declaration, config) =>
    config.tailwindConfig.corePlugins.alignSelf
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["align-self"])
      : [],

  animation: (declaration, config) =>
    config.tailwindConfig.corePlugins.animation
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.animation,
          classPrefix: "animate",
        })
      : [],

  appearance: (declaration, config) =>
    config.tailwindConfig.corePlugins.appearance
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["appearance"])
      : [],

  "aspect-ratio": (declaration, config) =>
    config.tailwindConfig.corePlugins.aspectRatio
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.aspectRatio,
          classPrefix: "aspect",
        })
      : [],

  "backdrop-filter": (declaration, config) => {
    if (!config.tailwindConfig.corePlugins.backdropFilter) {
      return [];
    }

    let classes: string[] = [];
    const mappings: Record<string, any> = {
      blur: config.tailwindConfig.corePlugins.backdropBlur && config.mapping.backdropBlur,
      brightness:
        config.tailwindConfig.corePlugins.backdropBrightness &&
        config.mapping.backdropBrightness,
      contrast:
        config.tailwindConfig.corePlugins.backdropContrast &&
        config.mapping.backdropContrast,
      grayscale:
        config.tailwindConfig.corePlugins.backdropGrayscale &&
        config.mapping.backdropGrayscale,
      "hue-rotate":
        config.tailwindConfig.corePlugins.backdropHueRotate &&
        config.mapping.backdropHueRotate,
      invert:
        config.tailwindConfig.corePlugins.backdropInvert && config.mapping.backdropInvert,
      opacity:
        config.tailwindConfig.corePlugins.backdropOpacity &&
        config.mapping.backdropOpacity,
      saturate:
        config.tailwindConfig.corePlugins.backdropSaturate &&
        config.mapping.backdropSaturate,
      sepia:
        config.tailwindConfig.corePlugins.backdropSepia && config.mapping.backdropSepia,
    };

    parseCSSFunctions(declaration.value).every(({ name, value }) => {
      if (name == null || value == null) {
        classes = [];
        return false;
      }

      const mapping = mappings[name];

      if (mapping) {
        delete mappings[name];

        const currentClasses = convertSizeDeclarationValue({
          declValue: value,
          valuesMap: mapping,
          classPrefix: `backdrop-${name}`,
          remInPx: config.remInPx,
          supportsNegativeValues: name === "hue-rotate",
        });

        if (currentClasses?.length) {
          classes = classes.concat(currentClasses);
          return true;
        }
      }

      classes = [];
      return false;
    });

    return classes;
  },

  "background-attachment": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundAttachment
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["background-attachment"]
        )
      : [],

  "background-blend-mode": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundBlendMode
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["background-blend-mode"]
        )
      : [],

  "background-clip": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundClip
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["background-clip"]
        )
      : [],

  "background-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.backgroundColor,
          classPrefix: "bg",
          cssDataType: "color",
        })
      : [],

  "background-image": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundImage
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.backgroundImage,
          classPrefix: "bg",
          fallbackValue: declaration.value,
          fallbackClassPrefix: "bg",
          cssDataType: "image",
        })
      : [],

  "background-origin": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundOrigin
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["background-origin"]
        )
      : [],

  "background-position": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundPosition
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.backgroundPosition,
          classPrefix: "bg",
          fallbackValue: declaration.value,
          fallbackClassPrefix: "bg",
          cssDataType: "position",
        })
      : [],

  "background-repeat": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundRepeat
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["background-repeat"]
        )
      : [],

  "background-size": (declaration, config) =>
    config.tailwindConfig.corePlugins.backgroundSize
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.backgroundSize,
          classPrefix: "bg",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  border: (declaration, config) =>
    convertBorderDeclarationValue({
      value: declaration.value,
      config,
      classPrefix: "border",
    }),

  "border-bottom": (declaration, config) =>
    convertBorderDeclarationValue({
      value: declaration.value,
      config,
      classPrefix: "border-b",
    }),

  "border-bottom-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderColor,
          classPrefix: "border-b",
          cssDataType: "color",
        })
      : [],

  "border-bottom-left-radius": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderRadius
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderRadius,
          classPrefix: "rounded-bl",
          remInPx: config.remInPx,
        })
      : [],

  "border-bottom-right-radius": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderRadius
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderRadius,
          classPrefix: "rounded-br",
          remInPx: config.remInPx,
        })
      : [],

  // 'border-bottom-style': (declaration, config) =>
  //   strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING['border-style']),

  "border-bottom-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderWidth,
          classPrefix: "border-b",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "border-collapse": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderCollapse
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["border-collapse"]
        )
      : [],

  "border-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderColor,
          classPrefix: "border",
          cssDataType: "color",
        })
      : [],

  "border-left": (declaration, config) =>
    convertBorderDeclarationValue({
      value: declaration.value,
      config,
      classPrefix: "border-l",
    }),

  "border-left-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderColor,
          classPrefix: "border-l",
          cssDataType: "color",
        })
      : [],

  // 'border-left-style': (declaration, config) =>
  //   strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING['border-style']),

  "border-left-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderWidth,
          classPrefix: "border-l",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "border-radius": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderRadius
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderRadius,
          classPrefix: "rounded",
          remInPx: config.remInPx,
        })
      : [],

  "border-right": (declaration, config) =>
    convertBorderDeclarationValue({
      value: declaration.value,
      config,
      classPrefix: "border-r",
    }),

  "border-right-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderColor,
          classPrefix: "border-r",
          cssDataType: "color",
        })
      : [],

  // 'border-right-style': (declaration, config) =>
  //   strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING['border-style']),

  "border-right-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderWidth,
          classPrefix: "border-r",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "border-spacing": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderSpacing
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderSpacing,
          classPrefix: "border-spacing",
          remInPx: config.remInPx,
        })
      : [],

  "border-style": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderStyle
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["border-style"]
        )
      : [],

  "border-top": (declaration, config) =>
    convertBorderDeclarationValue({
      value: declaration.value,
      config,
      classPrefix: "border-t",
    }),

  "border-top-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderColor,
          classPrefix: "border-t",
          cssDataType: "color",
        })
      : [],

  "border-top-left-radius": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderRadius
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderRadius,
          classPrefix: "rounded-tl",
          remInPx: config.remInPx,
        })
      : [],

  "border-top-right-radius": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderRadius
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderRadius,
          classPrefix: "rounded-tr",
          remInPx: config.remInPx,
        })
      : [],

  // 'border-top-style': (declaration, config) =>
  //   strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING['border-style']),

  "border-top-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderWidth,
          classPrefix: "border-t",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "border-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.borderWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.borderWidth,
          classPrefix: "border",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  bottom: (declaration, config) =>
    config.tailwindConfig.corePlugins.position
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.inset,
          classPrefix: "bottom",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "box-decoration-break": (declaration, config) =>
    config.tailwindConfig.corePlugins.position
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["box-decoration-break"]
        )
      : [],

  "box-shadow": (declaration, config) =>
    config.tailwindConfig.corePlugins.boxShadow
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.boxShadow,
          classPrefix: "shadow",
        })
      : [],

  "box-sizing": (declaration, config) =>
    config.tailwindConfig.corePlugins.boxSizing
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["box-sizing"])
      : [],

  "break-after": (declaration, config) =>
    config.tailwindConfig.corePlugins.breakAfter
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["break-after"])
      : [],

  "break-before": (declaration, config) =>
    config.tailwindConfig.corePlugins.breakBefore
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["break-before"]
        )
      : [],

  "break-inside": (declaration, config) =>
    config.tailwindConfig.corePlugins.breakInside
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["break-inside"]
        )
      : [],

  "caret-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.caretColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.caretColor,
          classPrefix: "caret",
          cssDataType: "color",
        })
      : [],

  clear: (declaration, config) =>
    config.tailwindConfig.corePlugins.clear
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["clear"])
      : [],

  color: (declaration, config) =>
    config.tailwindConfig.corePlugins.textColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.textColor,
          classPrefix: "text",
          cssDataType: "color",
        })
      : [],

  "column-gap": (declaration, config) =>
    config.tailwindConfig.corePlugins.gap
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.gap,
          classPrefix: "gap-x",
          remInPx: config.remInPx,
        })
      : [],

  columns: (declaration, config) =>
    config.tailwindConfig.corePlugins.columns
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.columns,
          classPrefix: "columns",
          remInPx: config.remInPx,
        })
      : [],

  content: (declaration, config) =>
    config.tailwindConfig.corePlugins.content
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.content,
          classPrefix: "content",
        })
      : [],

  cursor: (declaration, config) =>
    config.tailwindConfig.corePlugins.cursor
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.cursor,
          classPrefix: "cursor",
        })
      : [],

  display: (declaration, config) =>
    config.tailwindConfig.corePlugins.display
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["display"])
      : [],

  fill: (declaration, config) =>
    config.tailwindConfig.corePlugins.fill
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.fill,
          classPrefix: "fill",
        })
      : [],

  filter: (declaration, config) => {
    if (!config.tailwindConfig.corePlugins.filter) {
      return [];
    }

    let classes: string[] = [];
    const mappings: Record<string, any> = {
      blur: config.tailwindConfig.corePlugins.blur && config.mapping.blur,
      brightness:
        config.tailwindConfig.corePlugins.brightness && config.mapping.brightness,
      contrast: config.tailwindConfig.corePlugins.contrast && config.mapping.contrast,
      grayscale: config.tailwindConfig.corePlugins.grayscale && config.mapping.grayscale,
      "hue-rotate":
        config.tailwindConfig.corePlugins.hueRotate && config.mapping.hueRotate,
      invert: config.tailwindConfig.corePlugins.invert && config.mapping.invert,
      opacity: config.tailwindConfig.corePlugins.opacity && config.mapping.opacity,
      saturate: config.tailwindConfig.corePlugins.saturate && config.mapping.saturate,
      sepia: config.tailwindConfig.corePlugins.sepia && config.mapping.sepia,
      // 'drop-shadow': config.tailwindConfig.corePlugins.dropShadow && config.mapping.dropShadow,
    };

    parseCSSFunctions(declaration.value).every(({ name, value }) => {
      if (name == null || value == null) {
        classes = [];
        return false;
      }

      const mapping = mappings[name];

      if (mapping) {
        delete mapping[name];

        const currentClasses = convertSizeDeclarationValue({
          declValue: value,
          valuesMap: mapping,
          classPrefix: name,
          remInPx: config.remInPx,
          supportsNegativeValues: name === "hue-rotate",
        });

        if (currentClasses?.length) {
          classes = classes.concat(currentClasses);
          return true;
        }
      }

      classes = [];
      return false;
    });

    return classes;
  },

  flex: (declaration, config) => {
    if (!config.tailwindConfig.corePlugins.flex) {
      return [];
    }

    const classes = convertDeclarationValue({
      value: declaration.value,
      valuesMap: config.mapping.flex,
      classPrefix: "flex",
      fallbackValue: "",
    });

    if (classes.length) {
      return classes;
    }

    const [flexGrow, flexShrink = "1", flexBasis = "0%"] = declaration.value
      .trim()
      .split(/\s+/m);

    return convertDeclarationValue({
      value: `${flexGrow} ${flexShrink} ${flexBasis}`,
      valuesMap: config.mapping.flex,
      classPrefix: "flex",
      fallbackValue: declaration.value,
    });
  },

  "flex-basis": (declaration, config) =>
    config.tailwindConfig.corePlugins.flexBasis
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.flexBasis,
          classPrefix: "basis",
          remInPx: config.remInPx,
        })
      : [],

  "flex-direction": (declaration, config) =>
    config.tailwindConfig.corePlugins.flexDirection
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["flex-direction"]
        )
      : [],

  "flex-grow": (declaration, config) =>
    config.tailwindConfig.corePlugins.flexGrow
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.flexGrow,
          classPrefix: "grow",
        })
      : [],

  "flex-shrink": (declaration, config) =>
    config.tailwindConfig.corePlugins.flexShrink
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.flexShrink,
          classPrefix: "shrink",
        })
      : [],

  "flex-wrap": (declaration, config) =>
    config.tailwindConfig.corePlugins.flexWrap
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["flex-wrap"])
      : [],

  float: (declaration, config) =>
    config.tailwindConfig.corePlugins.float
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["float"])
      : [],

  "font-size": (declaration, config) =>
    config.tailwindConfig.corePlugins.fontSize
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.fontSize,
          classPrefix: "text",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "font-smoothing": (declaration, config) =>
    config.tailwindConfig.corePlugins.fontSmoothing
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["font-smoothing"]
        )
      : [],

  "font-style": (declaration, config) =>
    config.tailwindConfig.corePlugins.fontStyle
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["font-style"])
      : [],

  "font-variant-numeric": (declaration, config) =>
    config.tailwindConfig.corePlugins.fontVariantNumeric
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["font-variant-numeric"]
        )
      : [],

  "font-weight": (declaration, config) =>
    config.tailwindConfig.corePlugins.fontWeight
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.fontWeight,
          classPrefix: "font",
          fallbackValue: declaration.value,
          fallbackClassPrefix: "font",
          cssDataType: "number",
        })
      : [],

  gap: (declaration, config) =>
    config.tailwindConfig.corePlugins.gap
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.gap,
          classPrefix: "gap",
          remInPx: config.remInPx,
        })
      : [],

  "grid-auto-columns": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridAutoColumns
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridAutoColumns,
          classPrefix: "auto-cols",
        })
      : [],

  "grid-auto-flow": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridAutoFlow
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["grid-auto-flow"]
        )
      : [],

  "grid-auto-rows": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridAutoRows
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridAutoRows,
          classPrefix: "auto-rows",
        })
      : [],

  "grid-column": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridColumn
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridColumn,
          classPrefix: "col",
        })
      : [],

  "grid-column-end": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridColumnEnd
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridColumnEnd,
          classPrefix: "col-end",
        })
      : [],

  "grid-column-gap": (declaration, config) =>
    config.tailwindConfig.corePlugins.gap
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.gap,
          classPrefix: "gap-x",
          remInPx: config.remInPx,
        })
      : [],

  "grid-column-start": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridColumnStart
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridColumnStart,
          classPrefix: "col-start",
        })
      : [],

  "grid-gap": (declaration, config) =>
    config.tailwindConfig.corePlugins.gap
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.gap,
          classPrefix: "gap",
          remInPx: config.remInPx,
        })
      : [],

  "grid-row": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridRow
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridRow,
          classPrefix: "row",
        })
      : [],

  "grid-row-end": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridRowEnd
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridRowEnd,
          classPrefix: "row-end",
        })
      : [],

  "grid-row-gap": (declaration, config) =>
    config.tailwindConfig.corePlugins.gap
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.gap,
          classPrefix: "gap-y",
          remInPx: config.remInPx,
        })
      : [],

  "grid-row-start": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridRowStart
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridRowStart,
          classPrefix: "row-start",
        })
      : [],

  "grid-template-columns": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridTemplateColumns
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridTemplateColumns,
          classPrefix: "grid-cols",
        })
      : [],

  "grid-template-rows": (declaration, config) =>
    config.tailwindConfig.corePlugins.gridTemplateRows
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.gridTemplateRows,
          classPrefix: "grid-rows",
        })
      : [],

  height: (declaration, config) =>
    config.tailwindConfig.corePlugins.height
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.height,
          classPrefix: "h",
          remInPx: config.remInPx,
        })
      : [],

  inset: (declaration, config) =>
    config.tailwindConfig.corePlugins.inset
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.inset,
          classPrefix: "inset",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  isolation: (declaration, config) =>
    config.tailwindConfig.corePlugins.isolation
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["isolation"])
      : [],

  "justify-content": (declaration, config) =>
    config.tailwindConfig.corePlugins.justifyContent
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["justify-content"]
        )
      : [],

  "justify-items": (declaration, config) =>
    config.tailwindConfig.corePlugins.justifyItems
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["justify-items"]
        )
      : [],

  "justify-self": (declaration, config) =>
    config.tailwindConfig.corePlugins.justifySelf
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["justify-self"]
        )
      : [],

  left: (declaration, config) =>
    config.tailwindConfig.corePlugins.position
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.inset,
          classPrefix: "left",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "letter-spacing": (declaration, config) =>
    config.tailwindConfig.corePlugins.letterSpacing
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.letterSpacing,
          classPrefix: "tracking",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "line-height": (declaration, config) =>
    config.tailwindConfig.corePlugins.lineHeight
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.lineHeight,
          classPrefix: "leading",
          remInPx: config.remInPx,
        })
      : [],

  "list-style-position": (declaration, config) =>
    config.tailwindConfig.corePlugins.listStylePosition
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["list-style-position"]
        )
      : [],

  "list-style-type": (declaration, config) =>
    config.tailwindConfig.corePlugins.listStyleType
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.listStyleType,
          classPrefix: "list",
        })
      : [],

  margin: (declaration, config) =>
    config.tailwindConfig.corePlugins.margin
      ? convertComposedSpacingDeclarationValue(
          declaration.value,
          {
            top: { valuesMapping: config.mapping.margin, classPrefix: "mt" },
            right: { valuesMapping: config.mapping.margin, classPrefix: "mr" },
            bottom: { valuesMapping: config.mapping.margin, classPrefix: "mb" },
            left: { valuesMapping: config.mapping.margin, classPrefix: "ml" },
          },
          config.remInPx
        )
      : [],

  "margin-bottom": (declaration, config) =>
    config.tailwindConfig.corePlugins.margin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.margin,
          classPrefix: "mb",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "margin-left": (declaration, config) =>
    config.tailwindConfig.corePlugins.margin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.margin,
          classPrefix: "ml",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "margin-right": (declaration, config) =>
    config.tailwindConfig.corePlugins.margin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.margin,
          classPrefix: "mr",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "margin-top": (declaration, config) =>
    config.tailwindConfig.corePlugins.margin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.margin,
          classPrefix: "mt",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "max-height": (declaration, config) =>
    config.tailwindConfig.corePlugins.maxHeight
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.maxHeight,
          classPrefix: "max-h",
          remInPx: config.remInPx,
        })
      : [],

  "max-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.maxWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.maxWidth,
          classPrefix: "max-w",
          remInPx: config.remInPx,
        })
      : [],

  "min-height": (declaration, config) =>
    config.tailwindConfig.corePlugins.minHeight
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.minHeight,
          classPrefix: "min-h",
          remInPx: config.remInPx,
        })
      : [],

  "min-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.minWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.minWidth,
          classPrefix: "min-w",
          remInPx: config.remInPx,
        })
      : [],

  "mix-blend-mode": (declaration, config) =>
    config.tailwindConfig.corePlugins.mixBlendMode
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["mix-blend-mode"]
        )
      : [],

  "object-fit": (declaration, config) =>
    config.tailwindConfig.corePlugins.objectFit
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["object-fit"])
      : [],

  "object-position": (declaration, config) =>
    config.tailwindConfig.corePlugins.objectPosition
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.objectPosition,
          classPrefix: "object",
        })
      : [],

  opacity: (declaration, config) =>
    config.tailwindConfig.corePlugins.opacity
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.opacity,
          classPrefix: "opacity",
        })
      : [],

  order: (declaration, config) =>
    config.tailwindConfig.corePlugins.order
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.order,
          classPrefix: "order",
          remInPx: null,
          supportsNegativeValues: true,
        })
      : [],

  outline: (declaration, config) =>
    config.tailwindConfig.corePlugins.outlineStyle
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["outline"])
      : [],

  "outline-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.outlineColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.outlineColor,
          classPrefix: "outline",
          cssDataType: "color",
        })
      : [],

  "outline-offset": (declaration, config) =>
    config.tailwindConfig.corePlugins.outlineOffset
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.outlineOffset,
          classPrefix: "outline-offset",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
          cssDataType: "length",
        })
      : [],

  "outline-style": (declaration, config) =>
    config.tailwindConfig.corePlugins.outlineStyle
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["outline-style"]
        )
      : [],

  "outline-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.outlineWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.outlineWidth,
          classPrefix: "outline",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  overflow: (declaration, config) =>
    config.tailwindConfig.corePlugins.overflow
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["overflow"])
      : [],

  "overflow-wrap": (declaration, config) =>
    config.tailwindConfig.corePlugins.wordBreak
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["overflow-wrap"]
        )
      : [],

  "overflow-x": (declaration, config) =>
    config.tailwindConfig.corePlugins.overflow
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["overflow-x"])
      : [],

  "overflow-y": (declaration, config) =>
    config.tailwindConfig.corePlugins.overflow
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["overflow-y"])
      : [],

  "overscroll-behavior": (declaration, config) =>
    config.tailwindConfig.corePlugins.overscrollBehavior
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["overscroll-behavior"]
        )
      : [],
  "overscroll-behavior-x": (declaration, config) =>
    config.tailwindConfig.corePlugins.overscrollBehavior
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["overscroll-behavior-x"]
        )
      : [],
  "overscroll-behavior-y": (declaration, config) =>
    config.tailwindConfig.corePlugins.overscrollBehavior
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["overscroll-behavior-y"]
        )
      : [],

  padding: (declaration, config) =>
    config.tailwindConfig.corePlugins.padding
      ? convertComposedSpacingDeclarationValue(
          declaration.value,
          {
            top: { valuesMapping: config.mapping.padding, classPrefix: "pt" },
            right: { valuesMapping: config.mapping.padding, classPrefix: "pr" },
            bottom: {
              valuesMapping: config.mapping.padding,
              classPrefix: "pb",
            },
            left: { valuesMapping: config.mapping.padding, classPrefix: "pl" },
          },
          config.remInPx
        )
      : [],

  "padding-bottom": (declaration, config) =>
    config.tailwindConfig.corePlugins.padding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.padding,
          classPrefix: "pb",
          remInPx: config.remInPx,
        })
      : [],

  "padding-left": (declaration, config) =>
    config.tailwindConfig.corePlugins.padding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.padding,
          classPrefix: "pl",
          remInPx: config.remInPx,
        })
      : [],

  "padding-right": (declaration, config) =>
    config.tailwindConfig.corePlugins.padding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.padding,
          classPrefix: "pr",
          remInPx: config.remInPx,
        })
      : [],

  "padding-top": (declaration, config) =>
    config.tailwindConfig.corePlugins.padding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.padding,
          classPrefix: "pt",
          remInPx: config.remInPx,
        })
      : [],

  "page-break-after": (declaration, config) =>
    config.tailwindConfig.corePlugins.breakAfter
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["break-after"])
      : [],

  "page-break-before": (declaration, config) =>
    config.tailwindConfig.corePlugins.breakBefore
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["break-before"]
        )
      : [],

  "page-break-inside": (declaration, config) =>
    config.tailwindConfig.corePlugins.breakInside
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["break-inside"]
        )
      : [],

  "place-content": (declaration, config) =>
    config.tailwindConfig.corePlugins.placeContent
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["place-content"]
        )
      : [],

  "place-items": (declaration, config) =>
    config.tailwindConfig.corePlugins.placeItems
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["place-items"])
      : [],

  "place-self": (declaration, config) =>
    config.tailwindConfig.corePlugins.placeSelf
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["place-self"])
      : [],

  "pointer-events": (declaration, config) =>
    config.tailwindConfig.corePlugins.pointerEvents
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["pointer-events"]
        )
      : [],

  position: (declaration, config) =>
    config.tailwindConfig.corePlugins.position
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["position"])
      : [],

  resize: (declaration, config) =>
    config.tailwindConfig.corePlugins.resize
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["resize"])
      : [],

  right: (declaration, config) =>
    config.tailwindConfig.corePlugins.position
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.inset,
          classPrefix: "right",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "row-gap": (declaration, config) =>
    config.tailwindConfig.corePlugins.gap
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.gap,
          classPrefix: "gap-y",
          remInPx: config.remInPx,
        })
      : [],

  "scroll-behavior": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollBehavior
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["scroll-behavior"]
        )
      : [],

  "scroll-margin": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollMargin
      ? convertComposedSpacingDeclarationValue(
          declaration.value,
          {
            top: {
              valuesMapping: config.mapping.scrollMargin,
              classPrefix: "scroll-mt",
            },
            right: {
              valuesMapping: config.mapping.scrollMargin,
              classPrefix: "scroll-mr",
            },
            bottom: {
              valuesMapping: config.mapping.scrollMargin,
              classPrefix: "scroll-mb",
            },
            left: {
              valuesMapping: config.mapping.scrollMargin,
              classPrefix: "scroll-ml",
            },
          },
          config.remInPx
        )
      : [],

  "scroll-margin-bottom": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollMargin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollMargin,
          classPrefix: "scroll-mb",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "scroll-margin-left": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollMargin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollMargin,
          classPrefix: "scroll-ml",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "scroll-margin-right": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollMargin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollMargin,
          classPrefix: "scroll-mr",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "scroll-margin-top": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollMargin
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollMargin,
          classPrefix: "scroll-mt",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "scroll-padding": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollPadding
      ? convertComposedSpacingDeclarationValue(
          declaration.value,
          {
            top: {
              valuesMapping: config.mapping.scrollPadding,
              classPrefix: "scroll-pt",
            },
            right: {
              valuesMapping: config.mapping.scrollPadding,
              classPrefix: "scroll-pr",
            },
            bottom: {
              valuesMapping: config.mapping.scrollPadding,
              classPrefix: "scroll-pb",
            },
            left: {
              valuesMapping: config.mapping.scrollPadding,
              classPrefix: "scroll-pl",
            },
          },
          config.remInPx
        )
      : [],

  "scroll-padding-bottom": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollPadding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollPadding,
          classPrefix: "scroll-pb",
          remInPx: config.remInPx,
        })
      : [],

  "scroll-padding-left": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollPadding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollPadding,
          classPrefix: "scroll-pl",
          remInPx: config.remInPx,
        })
      : [],

  "scroll-padding-right": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollPadding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollPadding,
          classPrefix: "scroll-pr",
          remInPx: config.remInPx,
        })
      : [],

  "scroll-padding-top": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollPadding
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.scrollPadding,
          classPrefix: "scroll-pt",
          remInPx: config.remInPx,
        })
      : [],

  "scroll-snap-align": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollSnapAlign
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["scroll-snap-align"]
        )
      : [],

  "scroll-snap-type": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollSnapType
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["scroll-snap-type"]
        )
      : [],

  "scroll-snap-stop": (declaration, config) =>
    config.tailwindConfig.corePlugins.scrollSnapStop
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["scroll-snap-stop"]
        )
      : [],

  stroke: (declaration, config) =>
    config.tailwindConfig.corePlugins.stroke
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.stroke,
          classPrefix: "stroke",
          cssDataType: "color",
        })
      : [],

  "stroke-width": (declaration, config) =>
    config.tailwindConfig.corePlugins.strokeWidth
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.strokeWidth,
          classPrefix: "stroke",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "table-layout": (declaration, config) =>
    config.tailwindConfig.corePlugins.tableLayout
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["table-layout"]
        )
      : [],

  "text-align": (declaration, config) =>
    config.tailwindConfig.corePlugins.textAlign
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["text-align"])
      : [],

  "text-decoration": (declaration, config) => {
    if (!config.tailwindConfig.corePlugins.textDecoration) {
      return [];
    }

    const parsed = declaration.value.trim().split(/\s+/m);
    return parsed.length === 1
      ? strictConvertDeclarationValue(
          parsed[0],
          UTILITIES_MAPPING["text-decoration-line"]
        )
      : [];
  },

  "text-decoration-color": (declaration, config) =>
    config.tailwindConfig.corePlugins.textDecorationColor
      ? convertColorDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.textDecorationColor,
          classPrefix: "decoration",
          cssDataType: "color",
        })
      : [],

  "text-decoration-line": (declaration, config) =>
    config.tailwindConfig.corePlugins.textDecoration
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["text-decoration-line"]
        )
      : [],

  "text-decoration-style": (declaration, config) =>
    config.tailwindConfig.corePlugins.textDecorationStyle
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["text-decoration-style"]
        )
      : [],

  "text-decoration-thickness": (declaration, config) =>
    config.tailwindConfig.corePlugins.textDecorationThickness
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.textDecorationThickness,
          classPrefix: "decoration",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  "text-indent": (declaration, config) =>
    config.tailwindConfig.corePlugins.textIndent
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.textIndent,
          classPrefix: "indent",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "text-overflow": (declaration, config) =>
    config.tailwindConfig.corePlugins.textOverflow
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["text-overflow"]
        )
      : [],

  "text-transform": (declaration, config) =>
    config.tailwindConfig.corePlugins.textTransform
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["text-transform"]
        )
      : [],

  "text-underline-offset": (declaration, config) =>
    config.tailwindConfig.corePlugins.textUnderlineOffset
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.textUnderlineOffset,
          classPrefix: "underline-offset",
          remInPx: config.remInPx,
          supportsNegativeValues: false,
          cssDataType: "length",
        })
      : [],

  top: (declaration, config) =>
    config.tailwindConfig.corePlugins.position
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.inset,
          classPrefix: "top",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        })
      : [],

  "touch-action": (declaration, config) =>
    config.tailwindConfig.corePlugins.touchAction
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["touch-action"]
        )
      : [],

  transform: (declaration, config) => {
    if (!config.tailwindConfig.corePlugins.transform) {
      return [];
    }

    let classes: string[] = [];

    function convertTranslate(
      value: string,
      axis: "x" | "y" | "both" = "both"
    ): string[] {
      if (axis === "both") {
        const splitted = split(value);
        return splitted.length > 2
          ? []
          : [
              ...(splitted[0] ? convertTranslate(splitted[0], "x") : []),
              ...(splitted[1] ? convertTranslate(splitted[1], "y") : []),
            ];
      }

      return convertSizeDeclarationValue({
        declValue: value,
        valuesMap: config.mapping.translate,
        classPrefix: `translate-${axis}`,
        remInPx: config.remInPx,
        supportsNegativeValues: true,
      });
    }

    function convertSkew(value: string, axis: "x" | "y" | "both" = "both"): string[] {
      if (axis === "both") {
        const splitted = split(value);
        return splitted.length > 2
          ? []
          : [
              ...(splitted[0] ? convertSkew(splitted[0], "x") : []),
              ...(splitted[1] ? convertSkew(splitted[1], "y") : []),
            ];
      }

      return convertSizeDeclarationValue({
        declValue: value,
        valuesMap: config.mapping.skew,
        classPrefix: `skew-${axis}`,
        remInPx: config.remInPx,
        supportsNegativeValues: true,
      });
    }

    function convertScale(value: string, axis: "x" | "y" | "both" = "both"): string[] {
      if (axis === "both") {
        const splitted = split(value);

        if (splitted.length > 2) {
          return [];
        }

        if (splitted[0]) {
          return [
            ...convertScale(splitted[0], "x"),
            ...convertScale(splitted[1] || splitted[0], "y"),
          ];
        }
      }

      return convertSizeDeclarationValue({
        declValue: value,
        valuesMap: config.mapping.scale,
        classPrefix: `scale-${axis}`,
        remInPx: config.remInPx,
        supportsNegativeValues: true,
      });
    }

    parseCSSFunctions(declaration.value).every(({ name, value }) => {
      if (name == null || value == null) {
        classes = [];
        return false;
      }

      let converted: string[] = [];

      if (config.tailwindConfig.corePlugins.translate) {
        if (name === "translate") {
          converted = convertTranslate(value, "both");
        } else if (name === "translateX") {
          converted = convertTranslate(value, "x");
        } else if (name === "translateY") {
          converted = convertTranslate(value, "y");
        }
      }

      if (config.tailwindConfig.corePlugins.skew) {
        if (name === "skew") {
          converted = convertSkew(value, "both");
        } else if (name === "skewX") {
          converted = convertSkew(value, "x");
        } else if (name === "skewY") {
          converted = convertSkew(value, "y");
        }
      }

      if (config.tailwindConfig.corePlugins.scale) {
        if (name === "scale") {
          converted = convertScale(value, "both");
        } else if (name === "scaleX") {
          converted = convertScale(value, "x");
        } else if (name === "scaleY") {
          converted = convertScale(value, "y");
        }
      }

      if (config.tailwindConfig.corePlugins.rotate && name === "rotate") {
        converted = convertSizeDeclarationValue({
          declValue: value,
          valuesMap: config.mapping.rotate,
          classPrefix: "rotate",
          remInPx: config.remInPx,
          supportsNegativeValues: true,
        });
      }

      if (converted.length) {
        classes = classes.concat(converted);
        return true;
      }

      classes = [];
      return false;
    });

    return classes;
  },

  "transform-origin": (declaration, config) =>
    config.tailwindConfig.corePlugins.transformOrigin
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.transformOrigin,
          classPrefix: "origin",
        })
      : [],

  transition: (declaration, config) => {
    let classes: string[] = [];
    let hasDelay = false;

    removeUnnecessarySpaces(declaration.value.trim())
      .split(/\s+/m)
      .map(v => v.trim())
      .every((value, index) => {
        let itemClasses: string[] = [];

        if (index === 0) {
          itemClasses = config.tailwindConfig.corePlugins.transitionProperty
            ? convertDeclarationValue({
                value,
                valuesMap: config.mapping.transitionProperty,
                classPrefix: "transition",
              })
            : [];
        } else if (index === 1) {
          itemClasses = config.tailwindConfig.corePlugins.transitionDuration
            ? convertDeclarationValue({
                value,
                valuesMap: config.mapping.transitionDuration,
                classPrefix: "duration",
              })
            : [];
        } else if (index === 2) {
          const isTimingFunction = isNaN(parseFloat(value));

          if (isTimingFunction) {
            itemClasses = config.tailwindConfig.corePlugins.transitionTimingFunction
              ? convertDeclarationValue({
                  value,
                  valuesMap: config.mapping.transitionTimingFunction,
                  classPrefix: "ease",
                })
              : [];
          } else {
            hasDelay = true;
            itemClasses = config.tailwindConfig.corePlugins.transitionDelay
              ? convertDeclarationValue({
                  value,
                  valuesMap: config.mapping.transitionDelay,
                  classPrefix: "delay",
                })
              : [];
          }
        } else if (index === 3) {
          itemClasses =
            config.tailwindConfig.corePlugins.transitionDelay && !hasDelay
              ? convertDeclarationValue({
                  value,
                  valuesMap: config.mapping.transitionDelay,
                  classPrefix: "delay",
                })
              : [];

          hasDelay = true;
        }

        if (!itemClasses.length) {
          classes = [];
          return false;
        }

        classes = classes.concat(itemClasses);
        return true;
      });

    return classes;
  },

  "transition-delay": (declaration, config) =>
    config.tailwindConfig.corePlugins.transitionDelay
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.transitionDelay,
          classPrefix: "delay",
        })
      : [],

  "transition-duration": (declaration, config) =>
    config.tailwindConfig.corePlugins.transitionDuration
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.transitionDuration,
          classPrefix: "duration",
        })
      : [],

  "transition-property": (declaration, config) =>
    config.tailwindConfig.corePlugins.transitionProperty
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.transitionProperty,
          classPrefix: "transition",
        })
      : [],

  "transition-timing-function": (declaration, config) =>
    config.tailwindConfig.corePlugins.transitionTimingFunction
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.transitionTimingFunction,
          classPrefix: "ease",
        })
      : [],

  "user-select": (declaration, config) =>
    config.tailwindConfig.corePlugins.userSelect
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["user-select"])
      : [],

  "vertical-align": (declaration, config) =>
    config.tailwindConfig.corePlugins.verticalAlign
      ? strictConvertDeclarationValue(
          declaration.value,
          UTILITIES_MAPPING["vertical-align"]
        )
      : [],

  visibility: (declaration, config) =>
    config.tailwindConfig.corePlugins.visibility
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["visibility"])
      : [],

  "white-space": (declaration, config) =>
    config.tailwindConfig.corePlugins.whitespace
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["white-space"])
      : [],

  width: (declaration, config) =>
    config.tailwindConfig.corePlugins.width
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.width,
          classPrefix: "w",
          remInPx: config.remInPx,
        })
      : [],

  "will-change": (declaration, config) =>
    config.tailwindConfig.corePlugins.willChange
      ? convertDeclarationValue({
          value: declaration.value,
          valuesMap: config.mapping.willChange,
          classPrefix: "will-change",
        })
      : [],

  "word-break": (declaration, config) =>
    config.tailwindConfig.corePlugins.wordBreak
      ? strictConvertDeclarationValue(declaration.value, UTILITIES_MAPPING["word-break"])
      : [],

  "z-index": (declaration, config) =>
    config.tailwindConfig.corePlugins.zIndex
      ? convertSizeDeclarationValue({
          declValue: declaration.value,
          valuesMap: config.mapping.zIndex,
          classPrefix: "z",
          remInPx: null,
          supportsNegativeValues: true,
        })
      : [],
};

function split(value: string) {
  return value.split(/\s*,\s*/m).map(v => v.trim());
}
