import { colord } from "colord";
import type { Config } from "tailwindcss";
import type {
  KeyValuePair,
  RecursiveKeyValuePair,
  ScreensConfig,
} from "tailwindcss/types/config";
import type { ConverterMapping } from "../types";
import {
  buildMediaQueryByScreen,
  flattenObject,
  removeUnnecessarySpaces,
} from "../utils";

const remValueRegexp = /^(\d+)?\.?\d+rem$/;

function remValueToPx(value: string, remInPx: number) {
  if (remValueRegexp.test(value.trim())) {
    const number = parseFloat(value);

    return isNaN(number) ? value : `${number * remInPx}px`;
  }

  return value;
}

function normalizeNumbersInString(string: string) {
  return string.replace(/(^|[\s*+-/;])(\.\d+)/g, "$10$2");
}

export function normalizeValue(value: string) {
  return removeUnnecessarySpaces(normalizeNumbersInString(value));
}

export function normalizeColorValue(colorValue: string) {
  const parsed = colord(colorValue);
  return parsed.isValid() ? parsed.toHex() : colorValue;
}

export function normalizeZeroSizeValue(value: string) {
  return value.trim() === "0px" ? "0" : value;
}

export function normalizeSizeValue(
  sizeValue: string,
  remInPx: number | undefined | null
) {
  return normalizeZeroSizeValue(
    normalizeNumbersInString(
      remInPx != null ? remValueToPx(sizeValue, remInPx) : sizeValue
    )
  );
}

export function normalizeAtRuleParams(atRuleParam: string) {
  return removeUnnecessarySpaces(atRuleParam.replace(/\(|\)/g, ""));
}

function mapThemeTokens<V>(
  tokens: KeyValuePair<string, V>,
  valueConverterFn: (tokenValue: V, tokenKey: string) => string | null
) {
  const result: Record<string, string> = {};

  for (const [tokenKey, tokenValue] of Object.entries(tokens)) {
    const convertedTokenValue = valueConverterFn(tokenValue, tokenKey);

    if (convertedTokenValue) {
      result[convertedTokenValue] = tokenKey;
    }
  }

  return result;
}

function isColorKey(key: string) {
  return ["fill", "stroke"].includes(key) || key.toLowerCase().includes("color");
}

function isSizeKey(key: string) {
  return [
    "backdropBlur",
    "backgroundSize",
    "blur",
    "borderRadius",
    "borderSpacing",
    "borderWidth",
    "columns",
    "divideWidth",
    "flexBasis",
    "gap",
    "height",
    "inset",
    "letterSpacing",
    "lineHeight",
    "margin",
    "maxHeight",
    "maxWidth",
    "minHeight",
    "minWidth",
    "outlineOffset",
    "outlineWidth",
    "padding",
    "ringOffsetWidth",
    "ringWidth",
    "scrollMargin",
    "scrollPadding",
    "space",
    "spacing",
    "strokeWidth",
    "textDecorationThickness",
    "textUnderlineOffset",
    "translate",
    "width",
  ].includes(key);
}

function convertFontSizes(
  fontSizes: KeyValuePair<
    string,
    | string
    | [fontSize: string, lineHeight: string]
    | [
        fontSize: string,
        configuration: Partial<{
          lineHeight: string;
          letterSpacing: string;
          fontWeight: string | number;
        }>,
      ]
  >,
  remInPx?: number | null
) {
  return mapThemeTokens(fontSizes, fontSizeValue => {
    if (!fontSizeValue) {
      return null;
    }

    if (Array.isArray(fontSizeValue)) {
      fontSizeValue = fontSizeValue[0];
    }

    return normalizeSizeValue(fontSizeValue, remInPx);
  });
}

function convertScreens(screens: ScreensConfig) {
  if (Array.isArray(screens)) {
    return {} as Record<string, string>;
  }

  return mapThemeTokens(screens, screenValue =>
    screenValue ? normalizeAtRuleParams(buildMediaQueryByScreen(screenValue)) : null
  );
}

function convertColors(colors: RecursiveKeyValuePair) {
  const flatColors = flattenObject(colors);

  return mapThemeTokens(flatColors, (colorValue: string) => {
    colorValue = colorValue?.toString();

    return colorValue ? normalizeColorValue(colorValue) : null;
  });
}

function convertSizes(sizes: KeyValuePair, remInPx: number | null | undefined) {
  return mapThemeTokens(sizes, (sizeValue: string) => {
    sizeValue = sizeValue?.toString();

    return sizeValue ? normalizeSizeValue(sizeValue, remInPx) : null;
  });
}

function convertOtherThemeTokens(tokens: KeyValuePair | null | undefined) {
  return tokens
    ? mapThemeTokens(tokens, (tokenValue: string) => {
        tokenValue = tokenValue?.toString();

        return tokenValue ? normalizeValue(tokenValue) : null;
      })
    : tokens;
}

export function converterMappingByTailwindTheme(
  resolvedTailwindTheme: Config["theme"],
  remInPx?: number | null
) {
  const converterMapping = {} as ConverterMapping;

  if (!resolvedTailwindTheme) {
    return converterMapping;
  }

  for (const [key, themeItem] of Object.entries(
    resolvedTailwindTheme as Record<string, any>
  )) {
    if (["keyframes", "container", "fontFamily"].includes(key)) {
      continue;
    }

    if (key === "fontSize") {
      converterMapping[key] = convertFontSizes(themeItem, remInPx);
    } else if (key === "screens") {
      converterMapping[key] = convertScreens(themeItem);
    } else if (isColorKey(key)) {
      (converterMapping as any)[key] = convertColors(themeItem);
    } else if (isSizeKey(key)) {
      (converterMapping as any)[key] = convertSizes(themeItem, remInPx);
    } else {
      (converterMapping as any)[key] = convertOtherThemeTokens(themeItem);
    }
  }

  return converterMapping;
}
